from web3.auto import w3

from .tools.test_openzeppelin import TestOpenzeppelin
from src.predeployed_generator.openzeppelin.proxy_admin_generator import ProxyAdminGenerator
from src.predeployed_generator.openzeppelin.transparent_upgradeable_proxy_generator import TransparentUpgradeableProxyGenerator

class TestTransparentUpgradeableProxyGenerator(TestOpenzeppelin):
    OWNER_ADDRESS = '0xd200000000000000000000000000000000000000'
    PROXY_ADMIN_ADDRESS = '0xd200000000000000000000000000000000000001'
    PROXY_ADDRESS = '0xD200000000000000000000000000000000000002'
    IMPLEMENTATION_ADDRESS = '0xd200000000000000000000000000000000000003'


    def get_transparent_upgradeable_proxy_abi(self) -> list:
        return self.get_abi('TransparentUpgradeableProxy')


    def get_proxy_admin_abi(self) -> list:
        return self.get_abi('ProxyAdmin')

    def prepare_genesis(self):
        proxy_admin_generator = ProxyAdminGenerator()
        proxy_generator = TransparentUpgradeableProxyGenerator()

        genesis = self.generate_genesis({
            self.PROXY_ADMIN_ADDRESS: proxy_admin_generator.generate(owner_address=self.OWNER_ADDRESS),
            self.PROXY_ADDRESS: proxy_generator.generate(
                implementation_address=self.IMPLEMENTATION_ADDRESS,
                admin_address=self.PROXY_ADMIN_ADDRESS,
                initial_storage=proxy_admin_generator.generate_storage(owner_address=self.OWNER_ADDRESS)
            ),
            self.IMPLEMENTATION_ADDRESS: proxy_admin_generator.generate(owner_address=self.OWNER_ADDRESS)
        })
        return genesis

    # tests

    def test_admin(self, tmpdir):   
        self.datadir = tmpdir     
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()
            
            proxy_admin = w3.eth.contract(address=self.PROXY_ADMIN_ADDRESS, abi=self.get_proxy_admin_abi())
            assert proxy_admin.functions.getProxyAdmin(self.PROXY_ADDRESS).call() == self.PROXY_ADMIN_ADDRESS

    def test_implementation(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()
            
            proxy_admin = w3.eth.contract(address=self.PROXY_ADMIN_ADDRESS, abi=self.get_proxy_admin_abi())
            assert proxy_admin.functions.getProxyImplementation(self.PROXY_ADDRESS).call() == self.IMPLEMENTATION_ADDRESS 




'''access_control_enumerable_generator.py

Module for generation of contracts that are AccessControlEnumerable

classes:
    AccessControlEnumerableGenerator
'''

from dataclasses import dataclass
from typing import List

from ..contract_generator import ContractGenerator

class AccessControlEnumerableGenerator(ContractGenerator):
    '''Generates AccessControlEnumerable'''

    @dataclass
    class RolesSlots:
        '''Stores information about slots structure related to roles fields'''
        roles: int
        role_members: int

    # private

    @classmethod
    def _setup_role(
            cls,
            storage: dict,
            slots: RolesSlots,
            role: bytes,
            accounts: List[str]
            ) -> None:
        role_data_slot = cls.calculate_mapping_value_slot(
            slots.roles,
            role,
            'bytes32')
        members_slot = role_data_slot
        role_members_value_slot = cls.calculate_mapping_value_slot(
            slots.role_members,
            role,
            'bytes32')
        values_slot = role_members_value_slot
        indexes_slot = role_members_value_slot + 1
        cls._write_uint256(storage, values_slot, len(accounts))
        for i, account in enumerate(accounts):
            members_value_slot = cls.calculate_mapping_value_slot(
                members_slot,
                account,
                'address')
            cls._write_uint256(storage, members_value_slot, 1)
            cls._write_address(
                storage,
                cls.calculate_array_value_slot(values_slot, i),
                account)
            cls._write_uint256(
                storage,
                cls.calculate_mapping_value_slot(
                    indexes_slot,
                    int(account, 16),
                    'uint256'),
                i + 1)



"""upgradeable_contract_generator.py

The module is used to generate predeployed smart contracts in genesis block
that is upgradeable using EIP-1967 tranparent upgradeable proxy

Classes:
    UpgradeableContractGenerator
"""

from web3.auto import w3

from .openzeppelin.transparent_upgradeable_proxy_generator \
    import TransparentUpgradeableProxyGenerator
from .contract_generator import ContractGenerator


class UpgradeableContractGenerator(TransparentUpgradeableProxyGenerator):
    """Generates transparent upgradeable proxy based on implementation generator"""
    def __init__(
            self,
            implementation_generator: ContractGenerator):
        super().__init__()
        self.implementation_generator = implementation_generator

    def generate(self, balance=0, nonce=0, **_) -> ContractGenerator.Account:
        raise RuntimeError("""Can\'t generate upgradeable contract without implementation.
Use `generate_allocation` method instead""")

    def generate_allocation(
        self, contract_address, balance=0, nonce=0, **kwargs
    ) -> ContractGenerator.Allocation:
        """Generate smart contract allocation.
        It's pair of 2 smart contract:
        the first is upgradeable proxy
        and the second is an implementation

        If `implementation_address` parameter is not specified
        address of implementation is first 20 bytes of keccak256(`contract_address`)

        Arguments:
            - proxy_admin_address
            - arguments required by implementation contract generator

        Optional arguments:
            - implementation_address
            - balance
            - nonce

        Returns an object in format:
        {
            "0xd2...": {
                'balance': ... ,
                'nonce': ... ,
                'code': ... ,
                'storage': ...
            },
            "0xd2...": {
                'balance': ... ,
                'nonce': ... ,
                'code': ... ,
                'storage': ...
            }
        }
        """
        proxy_admin_address = kwargs.pop('proxy_admin_address')
        implementation_address = kwargs.pop(
            'implementation_address',
            w3.toChecksumAddress(w3.solidityKeccak(['address'], [contract_address])[:20]))

        return {
            contract_address: super().generate(
                admin_address=proxy_admin_address,
                implementation_address=implementation_address,
                initial_storage=self.implementation_generator.generate_storage(**kwargs),
                balance=balance,
                nonce=nonce),
            # pylint: disable=W0212
            implementation_address: self.implementation_generator._generate(storage=None)}



import pytest
from web3.auto import w3

from src.predeployed_generator.tools import MetaNotFoundError
from .tools.custom_contract_generator import CustomContractGenerator
from .tools.test_solidity_project import TestSolidityProject
from src.predeployed_generator.contract_generator import ContractGenerator


class TestContractGenerator(TestSolidityProject):
    OWNER_ADDRESS = '0xd200000000000000000000000000000000000000'
    TESTER_ADDRESS = '0xd200000000000000000000000000000000000001'
    TESTER2_ADDRESS = '0xD200000000000000000000000000000000000002'
    CONTRACT_ADDRESS = '0xd200000000000000000000000000000000000003'

    def get_test_contract_abi(self):
        return self.get_abi(CustomContractGenerator.CONTRACT_NAME)

    def prepare_genesis(self):
        test_contract_generator = CustomContractGenerator()

        return self.generate_genesis(test_contract_generator.generate_allocation(
            self.CONTRACT_ADDRESS,
            default_admin=self.OWNER_ADDRESS,
            testers=[self.TESTER_ADDRESS, self.TESTER2_ADDRESS]))

    def test_short_string(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()

            test_contract = w3.eth.contract(address=self.CONTRACT_ADDRESS, abi=self.get_test_contract_abi())
            assert test_contract.functions.shortString().call() == 'short string'

    def test_long_string(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()

            test_contract = w3.eth.contract(address=self.CONTRACT_ADDRESS, abi=self.get_test_contract_abi())
            assert test_contract.functions.longString().call() == ' '.join(['very'] * 32) + ' long string'

    def test_bytes32(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()

            test_contract = w3.eth.contract(address=self.CONTRACT_ADDRESS, abi=self.get_test_contract_abi())
            assert test_contract.functions.bytes32Value().call() == CustomContractGenerator.TESTER_ROLE

    def test_addresses_array(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()

            test_contract = w3.eth.contract(address=self.CONTRACT_ADDRESS, abi=self.get_test_contract_abi())
            assert test_contract.functions.testers(0).call() == self.TESTER_ADDRESS
            assert test_contract.functions.testers(1).call() == self.TESTER2_ADDRESS

    def test_generator_without_storage(self):
        class EmptyGenerator(ContractGenerator):
            pass

        bytecode = '0xbytecode'
        abi = ['function']
        meta = {'name': 'test'}
        balance = 5
        nonce = 13
        generator = EmptyGenerator(bytecode, abi, meta)
        assert generator.generate(balance=balance, nonce=nonce) == {
            'code': bytecode,
            'nonce': hex(nonce),
            'balance': hex(balance),
            'storage': {}
        }
        assert generator.get_abi() == abi
        assert generator.get_meta() == meta

    def test_non_existent_map_key_type(self):
        with pytest.raises(TypeError):
            ContractGenerator.calculate_mapping_value_slot(0, 'key', 'nonexistent')

    def test_generator_without_meta(self):
        class EmptyGenerator(ContractGenerator):
            pass

        bytecode = '0xbytecode'
        abi = ['function']
        generator = EmptyGenerator(bytecode, abi)
        assert generator.meta is None
        with pytest.raises(MetaNotFoundError):
            generator.get_meta()

    def test_generator_from_hardhat_artifact(self):
        class EmptyGenerator(ContractGenerator):
            pass

        generator = EmptyGenerator.from_hardhat_artifact(
            self.get_artifacts_path(CustomContractGenerator.CONTRACT_NAME)
        )
        assert generator.meta is None
        with pytest.raises(MetaNotFoundError):
            generator.get_meta()



'''
Creates ContractGenerator based on hardhat artifact file from `artifacts` folder

classes:
    OpenzeppelinContractGenerator
'''

import os

from ..contract_generator import ContractGenerator

class OpenzeppelinContractGenerator(ContractGenerator):
    '''Creates contract generator based on hardhat artifact file from `artifacts` folder'''
    ARTIFACT_FILENAME: str = ''
    META_FILENAME: str = ''

    def __init__(self):
        if not self.get_artifact_filename():
            raise TypeError('ARTIFACT_FILENAME is not overloaded in the class')
        if not self.get_meta_filename():
            raise TypeError('META_FILENAME is not overloaded in the class')
        artifacts_dir = os.path.join(os.path.dirname(__file__), 'artifacts')
        artifacts_path = os.path.join(artifacts_dir, self.get_artifact_filename())
        meta_path = os.path.join(artifacts_dir, self.get_meta_filename())
        contract = ContractGenerator.from_hardhat_artifact(artifacts_path, meta_path)
        super().__init__(bytecode=contract.bytecode, abi=contract.abi, meta=contract.meta)

    def get_artifact_filename(self):
        '''Return filename of contract artifact'''
        return self.ARTIFACT_FILENAME

    def get_meta_filename(self):
        '''Return filename of contract artifact'''
        return self.META_FILENAME



import json
import os

from src.predeployed_generator.tools import ArtifactsHandler
from test.tools.custom_contract_generator import CustomContractGenerator
from test.tools.test_solidity_project import TestSolidityProject


class TestArtifactsHandler(TestSolidityProject):

    def prepare(self):
        self.hardhat_dir = os.path.join(self.get_artifacts_dir(), 'contracts')
        self.target_dir = os.path.join(self.hardhat_dir, '..', 'target')
        os.mkdir(self.target_dir)
        self.name = CustomContractGenerator.CONTRACT_NAME

    def test_prepare_artifacts(self):
        self.prepare()
        test_handler = ArtifactsHandler(self.hardhat_dir, self.target_dir)
        test_handler.prepare_artifacts(self.name)

        assert os.path.exists(os.path.join(self.target_dir, f'{self.name}.json'))
        assert os.path.exists(os.path.join(self.target_dir, f'{self.name}.meta.json'))

        with open(os.path.join(self.target_dir, f'{self.name}.meta.json')) as meta_file:
            meta = json.loads(meta_file.read())
            assert meta['name'] == self.name



import pytest
from web3.auto import w3

from .tools.test_openzeppelin import TestOpenzeppelin
from src.predeployed_generator.openzeppelin.proxy_admin_generator import ProxyAdminGenerator
from src.predeployed_generator.openzeppelin.openzeppelin_contract_generator import OpenzeppelinContractGenerator

class TestProxyAdminGenerator(TestOpenzeppelin):
    def get_proxy_admin_abi(self) -> list:
        return self.get_abi('ProxyAdmin')


    def test_owner(self, tmpdir):
        self.datadir = tmpdir
        owner_address = '0xd200000000000000000000000000000000000000'
        proxy_admin_address = '0xd200000000000000000000000000000000000001'
        generator = ProxyAdminGenerator()
        with self.run_geth(tmpdir, self.generate_genesis({proxy_admin_address: generator.generate(owner_address=owner_address)})):
            assert w3.isConnected()
            
            proxy_admin = w3.eth.contract(address=proxy_admin_address, abi=self.get_proxy_admin_abi())
            assert proxy_admin.functions.owner().call() == owner_address

    def test_wrong_inheritance(self):
        class GeneratorWithoutArtifact(OpenzeppelinContractGenerator):
            pass

        class GeneratorWithoutMeta(OpenzeppelinContractGenerator):
            ARTIFACT_FILENAME = 'test'

        with pytest.raises(TypeError):
            GeneratorWithoutArtifact()

        with pytest.raises(TypeError):
            GeneratorWithoutMeta()



"""tools.py

The module stores helpers for work with predeployed contracts

Classes:
    ArtifactsHandler
"""

import json
import shutil
from os.path import normpath, join


class MetaNotFoundError(Exception):
    """Meta file does not exist"""


class ArtifactsHandler:
    """Generate and copy smart contract artifacts in pkg dir"""

    def __init__(self, hardhat_artifacts_dir, pkg_artifacts_dir):
        self.hardhat_artifacts_dir = hardhat_artifacts_dir
        self.pkg_artifacts_dir = pkg_artifacts_dir

    @staticmethod
    def get_build_info_path(contract_name, hardhat_contract_dir):
        """Get path to contract buildInfo"""
        with open(join(hardhat_contract_dir, f'{contract_name}.dbg.json'),
                  encoding='utf-8') as dbg_file:
            dbg = json.loads(dbg_file.read())
            return normpath(join(hardhat_contract_dir, dbg['buildInfo']))

    def get_hardhat_contract_dir(self, contract_name):
        """Get path to contract hardhat artifacts"""
        return normpath(join(self.hardhat_artifacts_dir, f'{contract_name}.sol'))

    def prepare_artifacts(self, contract_name):
        """Generates meta file and copy it and artifacts json to pkg dir"""
        hardhat_contract_dir = self.get_hardhat_contract_dir(contract_name)
        build_info_path = self.get_build_info_path(contract_name, hardhat_contract_dir)
        with open(build_info_path, encoding='utf-8') as info_file:
            info = json.loads(info_file.read())
            meta_data = {
                'name': contract_name,
                'solcVersion': info['solcVersion'],
                'solcLongVersion': info['solcLongVersion'],
                'input': info['input']
            }
        with open(join(self.pkg_artifacts_dir, f'{contract_name}.meta.json'),
                  mode='w',
                  encoding='utf-8') as meta:
            meta.write(json.dumps(meta_data, indent=4))
        shutil.copy(join(hardhat_contract_dir, f'{contract_name}.json'), self.pkg_artifacts_dir)



'''transparent_upgradeable_proxy_generator.py

Module for generation of transparent upgradeable proxy

classes:
    TransparentUpgradeableProxyGenerator
'''

from typing import Dict

from web3.auto import w3

from .openzeppelin_contract_generator import OpenzeppelinContractGenerator


class TransparentUpgradeableProxyGenerator(OpenzeppelinContractGenerator):
    '''Generates transparent upgradeable proxy
    '''
    ARTIFACT_FILENAME = 'TransparentUpgradeableProxy.json'
    META_FILENAME = 'TransparentUpgradeableProxy.meta.json'
    ROLLBACK_SLOT = int.from_bytes(
        w3.solidityKeccak(['string'], ['eip1967.proxy.rollback']),
        byteorder='big') - 1
    IMPLEMENTATION_SLOT = int.from_bytes(
        w3.solidityKeccak(['string'], ['eip1967.proxy.implementation']),
        byteorder='big') - 1
    ADMIN_SLOT = int.from_bytes(
        w3.solidityKeccak(['string'], ['eip1967.proxy.admin']),
        byteorder='big') - 1

    @staticmethod
    def generate_storage(**kwargs) -> dict:
        '''Generate contract storage

        Arguments:
            - implementation_address
            - admin_address

        Optional arguments:
            - initial_storage
        '''
        implementation_address = kwargs['implementation_address']
        admin_address = kwargs['admin_address']
        initial_storage = kwargs.get('initial_storage', None)
        storage: Dict[str, str] = {}
        TransparentUpgradeableProxyGenerator._write_address(
            storage,
            TransparentUpgradeableProxyGenerator.IMPLEMENTATION_SLOT,
            implementation_address)
        TransparentUpgradeableProxyGenerator._write_address(
            storage,
            TransparentUpgradeableProxyGenerator.ADMIN_SLOT,
            admin_address)
        if initial_storage is not None:
            storage.update(initial_storage)
        return storage



from web3.auto import w3

from .tools.custom_contract_generator import CustomContractGenerator
from .tools.test_solidity_project import TestSolidityProject


class TestAccessControlEnumerableGenerator(TestSolidityProject):
    OWNER_ADDRESS = '0xd200000000000000000000000000000000000000'
    TESTER_ADDRESS = '0xd200000000000000000000000000000000000001'
    TESTER2_ADDRESS = '0xD200000000000000000000000000000000000002'
    CONTRACT_ADDRESS = '0xd200000000000000000000000000000000000003'

    def get_test_contract_abi(self):
        return self.get_abi(CustomContractGenerator.CONTRACT_NAME)

    def prepare_genesis(self):
        test_contract_generator = CustomContractGenerator()

        return self.generate_genesis(test_contract_generator.generate_allocation(
            self.CONTRACT_ADDRESS,
            default_admin=self.OWNER_ADDRESS,
            testers=[self.TESTER_ADDRESS, self.TESTER2_ADDRESS]))
    
    def test_default_admin_role(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()

            test_contract = w3.eth.contract(address=self.CONTRACT_ADDRESS, abi=self.get_test_contract_abi())
            assert test_contract.functions.getRoleMemberCount(CustomContractGenerator.DEFAULT_ADMIN_ROLE).call() == 1
            assert test_contract.functions.getRoleMember(CustomContractGenerator.DEFAULT_ADMIN_ROLE, 0).call() == self.OWNER_ADDRESS            
            assert test_contract.functions.hasRole(CustomContractGenerator.DEFAULT_ADMIN_ROLE, self.OWNER_ADDRESS).call()

    def test_tester_role(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()

            test_contract = w3.eth.contract(address=self.CONTRACT_ADDRESS, abi=self.get_test_contract_abi())
            assert test_contract.functions.getRoleMemberCount(CustomContractGenerator.TESTER_ROLE).call() == 2
            assert test_contract.functions.getRoleMember(CustomContractGenerator.TESTER_ROLE, 0).call() == self.TESTER_ADDRESS
            assert test_contract.functions.getRoleMember(CustomContractGenerator.TESTER_ROLE, 1).call() == self.TESTER2_ADDRESS
            assert test_contract.functions.hasRole(CustomContractGenerator.TESTER_ROLE, self.TESTER_ADDRESS).call()
            assert test_contract.functions.hasRole(CustomContractGenerator.TESTER_ROLE, self.TESTER2_ADDRESS).call()



import pytest
from web3.auto import w3

from .tools.test_openzeppelin import TestOpenzeppelin
from src.predeployed_generator.openzeppelin.proxy_admin_generator import ProxyAdminGenerator
from src.predeployed_generator.openzeppelin.transparent_upgradeable_proxy_generator import TransparentUpgradeableProxyGenerator
from src.predeployed_generator.upgradeable_contract_generator import UpgradeableContractGenerator

class TestTransparentUpgradeableProxyGenerator(TestOpenzeppelin):
    OWNER_ADDRESS = '0xd200000000000000000000000000000000000000'
    PROXY_ADMIN_ADDRESS = '0xd200000000000000000000000000000000000001'
    PROXY_ADDRESS = '0xD200000000000000000000000000000000000002'
    PROXY_ADDRESS_HASH = '0xf4CD0343eb019A7869A0D28c8C3A546a313fEB51'
    IMPLEMENTATION_ADDRESS = '0xd200000000000000000000000000000000000003'


    def get_transparent_upgradeable_proxy_abi(self) -> list:
        return self.get_abi('TransparentUpgradeableProxy')

    def get_proxy_admin_abi(self) -> list:
        return self.get_abi('ProxyAdmin')

    def prepare_genesis(self):
        proxy_admin_generator = ProxyAdminGenerator()
        upgradeable_contract_generator = UpgradeableContractGenerator(proxy_admin_generator)

        return self.generate_genesis({
            self.PROXY_ADMIN_ADDRESS: proxy_admin_generator.generate(owner_address=self.OWNER_ADDRESS),
            **upgradeable_contract_generator.generate_allocation(
                contract_address=self.PROXY_ADDRESS,
                implementation_address=self.IMPLEMENTATION_ADDRESS,
                proxy_admin_address=self.PROXY_ADMIN_ADDRESS,
                owner_address=self.OWNER_ADDRESS
            )
        })

    # tests

    def test_admin(self, tmpdir):     
        self.datadir = tmpdir   
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()
            
            proxy_admin = w3.eth.contract(address=self.PROXY_ADMIN_ADDRESS, abi=self.get_proxy_admin_abi())
            assert proxy_admin.functions.getProxyAdmin(self.PROXY_ADDRESS).call() == self.PROXY_ADMIN_ADDRESS

    def test_implementation(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()
            
            proxy_admin = w3.eth.contract(address=self.PROXY_ADMIN_ADDRESS, abi=self.get_proxy_admin_abi())
            assert proxy_admin.functions.getProxyImplementation(self.PROXY_ADDRESS).call() == self.IMPLEMENTATION_ADDRESS

    def test_default_implementation_address(self, tmpdir):
        self.datadir = tmpdir
        proxy_admin_generator = ProxyAdminGenerator()
        upgradeable_contract_generator = UpgradeableContractGenerator(proxy_admin_generator)

        genesis = self.generate_genesis({
            self.PROXY_ADMIN_ADDRESS: proxy_admin_generator.generate(owner_address=self.OWNER_ADDRESS),
            **upgradeable_contract_generator.generate_allocation(
                contract_address=self.PROXY_ADDRESS,
                proxy_admin_address=self.PROXY_ADMIN_ADDRESS,
                owner_address=self.OWNER_ADDRESS
            )
        })

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()
            
            proxy_admin = w3.eth.contract(address=self.PROXY_ADMIN_ADDRESS, abi=self.get_proxy_admin_abi())
            assert proxy_admin.functions.getProxyImplementation(self.PROXY_ADDRESS).call() == self.PROXY_ADDRESS_HASH
    
    def test_owner(self, tmpdir):
        self.datadir = tmpdir
        genesis = self.prepare_genesis()

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()
            
            proxy_admin = w3.eth.contract(address=self.PROXY_ADDRESS, abi=self.get_proxy_admin_abi())
            assert proxy_admin.functions.owner().call() == self.OWNER_ADDRESS

    def test_wrong_generation(self):
        with pytest.raises(RuntimeError):
            contract_generator = UpgradeableContractGenerator(ProxyAdminGenerator())
            contract_generator.generate()

    def test_balance_and_nonce(self, tmpdir):
        self.datadir = tmpdir
        proxy_admin_generator = ProxyAdminGenerator()
        upgradeable_contract_generator = UpgradeableContractGenerator(proxy_admin_generator)

        balance = 5
        nonce = 1

        genesis = self.generate_genesis({
            self.PROXY_ADMIN_ADDRESS: proxy_admin_generator.generate(owner_address=self.OWNER_ADDRESS),
            **upgradeable_contract_generator.generate_allocation(
                balance=balance,
                nonce=nonce,
                contract_address=self.PROXY_ADDRESS,
                implementation_address=self.IMPLEMENTATION_ADDRESS,
                proxy_admin_address=self.PROXY_ADMIN_ADDRESS,
                owner_address=self.OWNER_ADDRESS
            )
        })

        with self.run_geth(tmpdir, genesis):
            assert w3.isConnected()
            
            assert w3.eth.get_balance(self.PROXY_ADDRESS) == balance
            assert w3.eth.get_transaction_count(self.PROXY_ADDRESS) == nonce




"""contract_generator.py

The module is used to generate predeployed smart contracts in genesis block

Functions:
    to_even_length
    add_0x
    calculate_mapping_value_slot
    calculate_array_value_slot
    next_slot

Classes:
    ContractGenerator
"""

from __future__ import annotations

import json
from typing import Dict, List, Union, Optional

from web3.auto import w3
from .tools import MetaNotFoundError


def to_even_length(hex_string: str) -> str:
    """Modify hex string to have even amount of digits.

    For example:
        "0x0" becomes "0x00"
        "0x123 becomes "0x0123"
        "0x12" stays "0x12"
    """
    assert hex_string.startswith('0x')
    if len(hex_string) % 2 != 0:
        return "0x0" + hex_string[2:]
    return hex_string


def add_0x(bytes_string: str) -> str:
    """Add "0x" prefix to the string"""
    if bytes_string.startswith('0x'):
        return bytes_string
    return '0x' + bytes_string


class ContractGenerator:
    """Generate smart contract allocation in a genesis block"""

    Storage = Dict[str, str]
    Account = Dict[str, Union[str, Storage]]
    Allocation = Dict[str, Account]

    def __init__(self, bytecode: str, abi: list, meta: Optional[dict] = None):
        self.bytecode = bytecode
        self.abi = abi
        self.meta = meta

    @staticmethod
    def from_hardhat_artifact(artifact_filename: str,
                              meta_filename: Optional[str] = None) -> ContractGenerator:
        """Create ContractGenerator from the artifact file built by hardhat"""
        with open(artifact_filename, encoding='utf-8') as artifact_file:
            contract = json.load(artifact_file)
        if meta_filename:
            with open(meta_filename, encoding='utf-8') as meta_file:
                meta = json.load(meta_file)
            return ContractGenerator(contract['deployedBytecode'], contract['abi'], meta)
        return ContractGenerator(contract['deployedBytecode'], contract['abi'])

    def generate(self, balance=0, nonce=0, **initial_values) -> Account:
        """Generate smart contract

        Returns an object in format:
        {
            'balance': ... ,
            'nonce': ... ,
            'code': ... ,
            'storage': ...
        }
        """
        return self._generate(self.generate_storage(**initial_values), balance, nonce)

    def generate_allocation(
        self, contract_address: str, balance=0, nonce=0, **args
    ) -> ContractGenerator.Allocation:
        """Generate smart contract allocation

        Returns an object in format:
        {
            "0xd2...": {
                'balance': ... ,
                'nonce': ... ,
                'code': ... ,
                'storage': ...
            }
        }
        """
        return {contract_address: self._generate(self.generate_storage(**args), balance, nonce)}

    @classmethod
    def generate_storage(cls, **_) -> Storage:
        """Generate smart contract storage layout
        based on initial values provided in args
        """
        return {}

    def get_abi(self) -> list:
        """Get the smart contract ABI
        """
        return self.abi

    def get_meta(self) -> dict:
        """Get the smart contract meta info
        """
        if not self.meta:
            raise MetaNotFoundError()
        return self.meta

    # private

    def _generate(self, storage: Storage = None, balance: int = 0, nonce: int = 0) -> Account:
        """Produce smart contract allocation object.

        It consists of fields 'code', 'balance', 'nonce' and 'storage'
        """
        assert isinstance(self.bytecode, str)
        assert isinstance(balance, int)
        assert isinstance(nonce, int)
        assert isinstance(storage, dict) or storage is None
        return {
            'code': self.bytecode,
            'balance': hex(balance),
            'nonce': hex(nonce),
            'storage': storage if storage is not None else {}
        }

    @staticmethod
    def _write_address(storage: Storage, slot: int, address: str) -> None:
        storage[to_even_length(hex(slot))] = address.lower()

    @staticmethod
    def _write_bytes32(storage: Storage, slot: int, data: bytes) -> None:
        assert len(data) <= 32
        storage[to_even_length(hex(slot))] = to_even_length(add_0x(data.hex()))

    @staticmethod
    def _write_uint256(storage: Storage, slot: int, value: int) -> None:
        storage[to_even_length(hex(slot))] = to_even_length(add_0x(hex(value)))

    @classmethod
    def _write_addresses_array(cls, storage: Storage, slot: int, values: List[str]) -> None:
        cls._write_uint256(storage, slot, len(values))
        for i, address in enumerate(values):
            address_slot = cls.calculate_array_value_slot(slot, i)
            cls._write_address(storage, address_slot, address)

    @classmethod
    def _write_string(cls, storage: Storage, slot: int, value: str) -> None:
        binary = value.encode()
        length = len(binary)
        if length < 32:
            binary += (2 * length).to_bytes(32 - length, 'big')
            cls._write_bytes32(storage, slot, binary)
        else:
            cls._write_uint256(storage, slot, 2 * length + 1)

            def chunks(size, source):
                for i in range(0, len(source), size):
                    yield source[i:i + size]

            for index, data in enumerate(chunks(32, binary)):
                if len(data) < 32:
                    data += int(0).to_bytes(32 - len(data), 'big')
                cls._write_bytes32(
                    storage,
                    cls.calculate_array_value_slot(slot, index),
                    data)

    @classmethod
    def calculate_mapping_value_slot(
            cls,
            slot: int,
            key: Union[bytes, int, str],
            key_type: str) -> int:
        """Calculate slot in smart contract storage where value of the key in mapping is stored"""

        if key_type == 'bytes32':
            assert isinstance(key, bytes)
        elif key_type == 'address':
            assert isinstance(key, str)
            return cls.calculate_mapping_value_slot(
                slot,
                int(key, 16).to_bytes(32, 'big'),
                'bytes32')
        elif key_type == 'uint256':
            assert isinstance(key, int)
        else:
            raise TypeError(f'{key_type} is unknown key type')

        return int.from_bytes(w3.solidityKeccak([key_type, 'uint256'], [key, slot]), 'big')

    @staticmethod
    def calculate_array_value_slot(slot: int, index: int) -> int:
        """Calculate slot in smart contract storage
        where value of the array in the index is stored
        """
        return int.from_bytes(w3.solidityKeccak(['uint256'], [slot]), 'big') + index

    @staticmethod
    def next_slot(previous_slot: int) -> int:
        """Return next slot in smart contract storage"""
        return previous_slot + 1



'''proxy_admin_generator.py

Module for generation of proxy admin

classes:
    ProxyAdminGenerator
'''

from typing import Dict

from .openzeppelin_contract_generator import OpenzeppelinContractGenerator


class ProxyAdminGenerator(OpenzeppelinContractGenerator):
    '''Generates ProxyAdmin
    '''

    ARTIFACT_FILENAME = 'ProxyAdmin.json'
    META_FILENAME = 'ProxyAdmin.meta.json'

    OWNER_SLOT = 0

    # ---------- storage ----------
    # -----------Context-----------
    # -----------Ownable-----------
    # 0:    _owner
    # ---------ProxyAdmin----------

    @staticmethod
    def generate_storage(**kwargs) -> Dict[str, str]:
        owner_address = kwargs['owner_address']
        storage: Dict[str, str] = {}
        ProxyAdminGenerator._write_address(
            storage,
            ProxyAdminGenerator.OWNER_SLOT,
            owner_address)
        return storage
